# FIXME: the following import should be done with a zope interfaces registry,
# and the Zope component architecture, since  the classes have already been
# registered as ITranscoder.

from . import basic, gstreamerffmpeg  # @UnusedImport
